/* graph search algorithm for AGRAPH */

#ifndef _agrpalg_c_
#define _agrpalg_c_

#include"agrhalg.h"
#include"queue.c"
#include"agraph.c"

/* print search tree */
void AGRAPH_search_tree_print (AGRAPH *G, ALIST_ID *vf, ALIST_ID *dep){
  ALIST_ID i;
  FLOOP (i, 0, G->node_end){
    if ( vf[i]>=G->edge_end*2 ) printf ("(" ALIST_IDF ":***) ", i); 
    else if ( vf[i]>=0 )
      if ( dep ) printf ("(" ALIST_IDF "->" ALIST_IDF "[" ALIST_IDF "]) ", i, AGRAPH_INC(*G,vf[i]),dep[i]); 
      else  printf ("(" ALIST_IDF "->" ALIST_IDF ") ", i, AGRAPH_INC(*G,vf[i]));
    else printf ("(" ALIST_IDF "->#" ALIST_IDF ") ", i, vf[i]); 
  }
  printf ("\n");
}


/* graph search starting from v. use edges. without recursion 
 Q is the queue storing visited vertices, vf is the internal edges to go to parent
  of each vertex. it is minus number when un-visited. */
/* if v is un-visited, node flag has to be minus number. graph search can 
 start even if its node flag is "visited" */
/* dep memories the distance from v on the search tree. */
/* flag &1 ==0 then depth first search. otherwise breadth-first */
ALIST_ID *AGRAPH_search (AGRAPH *G, ALIST_ID v, ALIST_ID *vf, ALIST_ID *dep, QUEUE *Q, int flag){
  ALIST_ID u, e=MALIST_HEAD (G->edge, v), ff=0, tt;
  QUEUE R;
  if ( vf == NULL ){
    malloc2 (vf, G->node_end, EXIT0);
    ARY_FILL (vf, 0, G->node_end, -1);
  }
  if ( Q == NULL ){ /* if Q is not initialized, initialize it. */
    ff=-1;
    Q = &R;
    QUEUE_alloc (Q, G->edge_end);
  } else {
    ff = Q->s;
    Q->s = Q->t;
  }
  tt = Q->t;
  if ( vf[v] < 0 ) vf[v]=G->edge_end*2; /* if root is un-visited, put special mark */
  if ( dep ) if ( dep[v]<0 ) dep[v]=0;
  while (1){
    MALIST_DO_FORWARD (G->edge, v, e){ /* insert edges incident to v */
      u = AGRAPH_INC (*G, e);
      if ( vf[u] < 0 ){
        QUE_INS (*Q, e);
        if ( tt < Q->t ) tt = Q->t;
        if ( (flag&3)==1 ) vf[u] = G->node_end*3;
      }
    }
    LOOP:;
    if ( QUEUE_LENGTH_(*Q) == 0 ) break;
    if ( (flag&3)!=0 ) e = QUEUE_ext_ (Q);
    else e = QUEUE_ext_tail (Q);
    v = AGRAPH_INC (*G, e);
    if ( vf[v] >= 0 && (flag&3)!=1 ) goto LOOP;
    vf[v] = AGRAPH_OPO(*G,e);
    if ( dep ) dep[v] = dep[AGRAPH_INC(*G,vf[v])]+1;
  }
  if ( ff == -1 ) QUEUE_end (Q);
  else {
    Q->s = ff;
    Q->t = tt;
  }
  return (vf);
}



/* initialize the flags and marks for graph search */
void AGRAPH_search_init (AGRAPH *G, ALIST_ID **vf, QUEUE *ff, ALIST_ID **dep){
  malloc2 (*vf, G->node_end, EXIT);
  ARY_FILL (*vf, 0, G->node_end, -1);
  malloc2 (*dep, G->node_end, EXIT0);
  ARY_FILL (*dep, 0, G->node_end, -1);
  QUEUE_alloc (ff, G->node_end+1);
}

/* initialize the marks where the graph search visited. */
void AGRAPH_search_reinit (ALIST_ID *vf, QUEUE *ff){
  QUEUE_INT *x;
  MQUE_FLOOP (*ff, x) vf[*x] = -1;
  ff->t = ff->s;
}

/* termination */
void AGRAPH_search_end (ALIST_ID *vf, QUEUE *ff, int *dep){
  mfree (vf, dep);
  QUEUE_end (ff);
}

/* v Ƃ[DTBċA^ł͂Ȃ
 flag  0 ȂΖAȂLAȂtɗL}ǂ
 vf m[h̃tOBSāiTōsƂj-1 ɂĂ
 ff ̓}[N_̗B}[NƂɎg
 TIAe_ɂ͐e֍s}̊ǗԍAA[N̔ԍ
 tɁA炩߃m[hɈĂ΁A͒TȂ */
void AGRAPH_DFS (AGRAPH *G, ALIST_ID v, ALIST_ID *vf, QUEUE *ff, int flag){
  ALIST_ID u, e;
  MALIST *L = AGRAPH_EDGES(*G,flag);
  e = MALIST_HEAD (*L, v);
   /* Ƀ}[NĂȂΕtB̐e͂Ȃ̂,݂Ȃ}w */
  if ( vf[v] < 0 ) vf[v] = L->end*2;
  QUE_INS (*ff, v);
  while (1){
    if ( MALIST_IS_ELM (*L, e) ){
      u = AGRAPH_INC_TO(*G,e,flag);  /* uɖK悤Ƃ钸_ */
      if ( vf[u] < 0 ){   /* u Ƀ}[NĂȂΖK */
        vf[u] = e;  /* ũ}[NɁAǂĂ}L^ */
        QUEUE_ins_ (ff, u);  /* uT_L[ɑ} */
        e = MALIST_HEAD (*L, u);
      } else {
        e = L->nxt[e];  /* }[NĂA̎}ɍs */
      }
    } else { /* KĂ钸_̎}S */
      u = MALIST_LIST (*L,e);
      if ( u == v ) break;  /* AI */  
      e = vf[u];
    }
  }
}

/* OtǴAv ƂLEOt̕DT
 flag  0 ȂΖAȊOȂL}ǂ
 vf m[h̃tOBSāiTōsƂj-1 ɂĂ
 ff ̓}[N_̗BASYŃL[ƂĎgB
 }[NƂɂgB
 dep ɂ́A[Bdep[v] v ̐[ĂƁBe_̐[́A
 dep[v] + [v ̋] ƂȂ */
/* TIAe_ɂ͐e֍s}̊ǗԍAA[N̔ԍ
 tɁA炩߃m[hɈĂ΁A͒TȂ */
/* X^[gn_̋ ub 𒴂TȂ */
void AGRAPH_BFS (AGRAPH *G, ALIST_ID v, ALIST_ID *vf, QUEUE *ff, ALIST_ID *dep, int flag, int ub){
  ALIST_ID u, e, s=ff->t, d, vv;

  MALIST *L = AGRAPH_EDGES(*G,flag);
  e = MALIST_HEAD (*L, v);
   /* Ƀ}[NĂȂΕtB̐e͂Ȃ̂,݂Ȃ}w */
  if ( vf[v] < 0 ) vf[v] = L->end*2;
  QUE_INS (*ff, v);
  while (s < ff->t){
    vv = ff->v[s];
    s++;
    if ( (d = dep[vv]) >= ub ) break;
    MALIST_DO_FORWARD (*L, vv, e){
      u = AGRAPH_INC_TO(*G,e,flag);  /* uɖK悤Ƃ钸_ */
      if ( vf[u] < 0 ){   /* u Ƀ}[NĂȂΖK */
        vf[u] = e;  /* ũ}[NɁAǂĂ}L^ */
        QUE_INS (*ff, u);  /* uT_L[ɑ} */
        dep[u] = d+1;
      }
    }
  }
}

/* v Ƃ[DTB_ɒT}[NtBmark }[N̔z
 Ep&1 Ȃ揇A p&2 Ȃ㏇Ap=4 Ȃ, 
 EOtŁAp=5 Ȃ揇ŁATCNɊ܂܂Ȃ tree edge  vf 
    G->edge_end+G->arc_end ă}[NtB
 Ep&8 Ȃ΁A}}[NBtree edge Ȃ0AobN}Ȃ1AtH[hANX}Ȃ2
 E}[N0n܂B*/
void AGRAPH_DFS_node_mark (AGRAPH *G, ALIST_ID v, ALIST_ID *vf, QUEUE *ff, int flag, ALIST_ID *mark, ALIST_ID p){
  ALIST_ID u, e, c=0, d=0, ee, M = G->edge_end + G->arc_end;
  MALIST *L = AGRAPH_EDGES(*G,flag);
  e = MALIST_HEAD (*L, v);
    /* Ƀ}[NĂȂΕtB̐e͂Ȃ̂,݂Ȃ}w */
  if ( vf[v] < 0 ) vf[v] = L->end*2;
  QUE_INS (*ff, v);
  if ( p&1 || p==4 ) mark[v] = c++;  /* 揇}[N */
  while (1){
    if ( MALIST_IS_ELM (*L, e) ){
      u = AGRAPH_INC_TO(*G,e,flag);  /* uɖK悤Ƃ钸_ */
      if ( vf[u] < 0 ){   /* u Ƀ}[NĂȂΖK */
        vf[u] = e + M;  /* ũ}[NɁAǂĂ}L^ */
        QUE_INS (*ff, u);  /* uT_L[ɑ} */
        if ( p&1 || (p==4&&((++d)%2==0)) ) mark[u] = c++;  /* 揇}[N */
        if ( p&8 ) mark[e] = 0;
        e = MALIST_HEAD (*L, u);
      } else {
        e = L->nxt[e];  /* }[NĂA̎}ɍs */
        if ( p&8 ) mark[e] = vf[u]<M? 2: 1;
      }
    } else { /* KĂ钸_̎}S */
      u = MALIST_LIST (*L,e);
      vf[u] -= M; e = vf[u];
      if ( p&2 || (p==4&&((++d)%2==0)) ) mark[u] = c++;   /* ㏇}[N */
      if ( p==5 ){  /* ̎q炢Ƃ̒_̓YL^ */
        MALIST_DO_FORWARD (G->edge, u, ee){ 
//          printf ("%d %d\n", u, ee );
          d = mark[AGRAPH_INC_TO(*G,ee,flag)];
          if ( mark[u]>d && (flag || e!=AGRAPH_OPO(*G,ee)) ){
            mark[u] = d;
            if ( vf[u] < M ) vf[u] += M;
          }
        }
      }
      if ( u == v ) break;  /* AI */  
    }
  }
}








/* Ep&8 Ȃ΁A}}[NBtree edge Ȃ0AobNȂ1ANXȂ2
    tH[hȂ3
*/










/*
main (){
  int *vf, i, *dep, *mark;
  QUEUE ff;
  AGRAPH G;
//  AGRAPH_init ( &G, 10, 500, 0 );
//  G.node1_num = 0;
//  AGRAPH_mk_rand_graph1 ( &G, 500, 0 );

  AGRAPH_load ( &G, "G_.GRP", 1, 0, 0, 0, 0);

//  AGRAPH_print ( &G );
//  AGRAPH_node_out ( &G, 0, 0 );
//  AGRAPH_node_out ( &G, 1, 0 );
//  AGRAPH_print ( &G );
//  AGRAPH_node_recov ( &G, 1, 0 );
//  AGRAPH_node_recov ( &G, 0, 0 );
//  AGRAPH_print ( &G );

  AGRAPH_search_init ( &G, &vf, &ff, &dep );
  mark = intarray_malloc_const ( G.node_end, -1 );
  AGRAPH_DFS_node_mark ( &G, 0, vf, &ff, 0, mark, 5 );
  for ( i=0 ; i<10 ; i++ ){
    if ( vf[i] > G.edge_end + G.arc_end ) printf ("*"); else printf (" ");
    printf ("%d:(%d %d) %d\n", i, vf[i], G.edge.list[vf[i]], mark[i]);
//    printf ("%d:(%d %d)\n", i, vf[i], AGRAPH_FROM(G, vf[i]));
  }
  AGRAPH_search_end ( vf, &ff, dep );
}
*/

#endif


